import numpy as np
from scipy.integrate import solve_ivp

import openmdao.api as om

from ..._options import options as dymos_options

from .ode_evaluation_group import ODEEvaluationGroup
from dymos.utils.misc import create_subprob, om_version


class ODEIntegrationComp(om.ExplicitComponent):
    """
    A component to perform explicit integration with a generic ODE integrator/IVP solver.

    This component contains a sub-Problem with a component that will be solved over num_nodes
    points instead of creating num_nodes instances of that same component and connecting them
    together.

    Parameters
    ----------
    input_grid_data : GridData
        The GridData which defines the nodes at which the control inputs values are specified.
    time_options : OptionsDictionary
        OptionsDictionary of time options.
    state_options : dict of {str: OptionsDictionary}
        For each state variable, a dictionary of its options, keyed by name.
    parameter_options : dict of {str: OptionsDictionary}
        For each parameter, a dictionary of its options, keyed by name.
    control_options : dict of {str: OptionsDictionary}
        For each control variable, a dictionary of its options, keyed by name.
    output_grid_data : GridData
        The GridData which defines the nodes at which the outputs of the integration are provided, or None if
        the input_grid_data is to be used.
    reports : bool or None or str or Sequence
        Controls the reports generated by the subproblems used during the integration.
    standalone_mode : bool
        If True, assume this component is being run as its own system. As part of the Dymos
        ShootingPhase, this system needs to be setup during the Phase configure process and setting this
        to False will enable that behavior.
    calc_exprs : dict
        ODE Expressions provided by the controlling phase.
    **kwargs : dict
        Additional keyword arguments passed to Group.

    Notes
    -----
    This code includes the following unicode symbols:
    theta:  U+03B8
    """
    def __init__(self, input_grid_data, time_options, state_options, parameter_options=None, control_options=None,
                 output_grid_data=None, reports=False, standalone_mode=True, calc_exprs=None, **kwargs):
        super().__init__(**kwargs)
        self.time_options = time_options
        self.state_options = state_options
        self.parameter_options = parameter_options or {}
        self.control_options = control_options or {}
        self._eval_subprob = None
        self._input_grid_data = input_grid_data
        self._output_grid_data = output_grid_data if output_grid_data is not None else input_grid_data
        self._reports = reports
        self._standalone_mode = standalone_mode

        self._inputs_cache = ''

        self.x_size = 0
        self.p_size = 0
        self.u_size = 0
        self.up_size = 0
        self.theta_size = 0
        self.z_size = 0

        self._state_rate_of_names = []
        self._totals_of_names = []
        self._totals_wrt_names = []

        self._no_check_partials = not dymos_options['include_check_partials']
        self._num_control_input_nodes = input_grid_data.subset_num_nodes['control_input']

        self._calc_exprs = {} if calc_exprs is None else calc_exprs

    def initialize(self):
        """
        Declare options for the ODEIntegrationComp.
        """
        self.options.declare('ode_class', desc='System defining the ODE', recordable=False)
        self.options.declare('method', default='DOP853', desc='The integration method used.')
        self.options.declare('atol', types=float, default=1.0E-6)
        self.options.declare('rtol', types=float, default=1.0E-9)
        self.options.declare('first_step', types=float, allow_none=True, default=None)
        self.options.declare('max_step', types=float, default=np.inf)
        self.options.declare('propagate_derivs', types=bool, default=True,
                             desc='If True, propagate the state and derivatives of the state and time with respect to '
                                  'the integration parameters. If False, only propagate the primal states.')
        self.options.declare('ode_init_kwargs', types=dict, allow_none=True, default=None)
        self.options.declare('control_interp', values=['vandermonde', 'barycentric', 'cubic'], default='vandermonde',
                             desc='Control interpolation algorithm, one of either "vandermonde" or "barycentric". In '
                             'general, Vandermonde is faster but Barycentric is necessary for the Birkhoff '
                             'transcription where the number of nodes per segment can exceed 20 to 30.')

    def _setup_subprob(self):
        self._eval_subprob = p = create_subprob(base_name=f'{self.pathname}_subprob',
                                                comm=self.comm,
                                                reports=self._reports)

        p.model.add_subsystem('ode_eval',
                              ODEEvaluationGroup(ode_class=self.options['ode_class'],
                                                 time_options=self.time_options,
                                                 state_options=self.state_options,
                                                 parameter_options=self.parameter_options,
                                                 control_options=self.control_options,
                                                 ode_init_kwargs=self.options['ode_init_kwargs'],
                                                 input_grid_data=self._input_grid_data,
                                                 compute_derivs=self.options['propagate_derivs'],
                                                 control_interp=self.options['control_interp'],
                                                 calc_exprs=self._calc_exprs),
                              promotes_inputs=['*'],
                              promotes_outputs=['*'])

        if om_version()[0] <= (3, 34, 2):
            p.setup(check=None)
        else:
            p.setup(check=None, parent=self)
        p.final_setup()

    def _set_segment_index(self, idx):
        """
        Set the index of the segment being integrated.
        """
        self._eval_subprob.model._get_subsystem('ode_eval').set_segment_index(idx)

    def _setup_time(self):
        if self._standalone_mode:
            self._configure_time()

    def _configure_time(self):
        """
        Components do not have configure methods, but since we rely on configure-time introspection to determine
        properties of the states, times, controls, parameters, and timeseries, we need to call this method at
        configure time in the parent ExplicitShooting transcription object.
        """
        num_output_rows = self._num_output_rows
        t_units = self.time_options['units']
        t_name = self.time_options['name']

        self._totals_of_names.append(t_name)
        self._totals_wrt_names.extend([t_name, 't_initial', 't_duration'])

        self.add_input('t_initial', shape=(1,), units=t_units)
        self.add_input('t_duration', shape=(1,), units=t_units)
        self.add_output('t_final', shape=(1,), units=t_units)
        self.add_output(t_name, shape=(num_output_rows, 1), units=t_units)
        self.add_output(f'{t_name}_phase', shape=(num_output_rows, 1), units=t_units)

        self.declare_partials('t_final', ['t_initial', 't_duration'], val=1.0)
        self.declare_partials(t_name, ['t_initial', 't_duration'], val=1.0)
        self.declare_partials(f'{t_name}_phase', 't_duration', val=1.0)

    def _setup_states(self):
        if self._standalone_mode:
            self._configure_states()

    def _configure_states(self):
        """
        Components do not have configure methods, but since we rely on configure-time introspection to determine
        properties of the states, times, controls, parameters, and timeseries, we need to call this method at
        configure time in the parent ExplicitShooting transcription object.
        """
        num_output_rows = self._num_output_rows

        # The total size of the entire state vector
        self.x_size = 0
        self.state_sizes = {}

        self._state_input_names = {}
        self._state_output_names = {}

        # The indices of each state in x
        self.state_idxs = {}

        # The indices of each state's initial value in z
        self._state_idxs_in_z = {}

        for state_name, options in self.state_options.items():
            self._state_input_names[state_name] = f'states:{state_name}'
            self._state_output_names[state_name] = f'states_out:{state_name}'

            # Keep track of the derivative "of" names for state rates separately, so we don't
            # request them when they're not necessary.
            self._state_rate_of_names.append(f'state_rate_collector.state_rates:{state_name}_rate')
            self._totals_wrt_names.append(self._state_input_names[state_name])

            self.add_input(self._state_input_names[state_name],
                           shape=options['shape'],
                           units=options['units'],
                           desc=f'initial value of state {state_name}')
            self.add_output(self._state_output_names[state_name],
                            shape=(num_output_rows,) + options['shape'],
                            units=options['units'],
                            desc=f'final value of state {state_name}')

            self.state_sizes[state_name] = state_size = np.prod(options['shape'], dtype=int)

            # The indices of the state in x
            self.state_idxs[state_name] = np.s_[self.x_size:self.x_size + state_size]
            self.x_size += state_size

            self.declare_partials(of=self._state_output_names[state_name],
                                  wrt='t_initial')

            self.declare_partials(of=self._state_output_names[state_name],
                                  wrt='t_duration')

            for state_name_wrt in self.state_options:
                self.declare_partials(of=self._state_output_names[state_name],
                                      wrt=f'states:{state_name_wrt}')

            for param_name_wrt in self.parameter_options:
                self.declare_partials(of=self._state_output_names[state_name],
                                      wrt=f'parameters:{param_name_wrt}')

            for control_name_wrt in self.control_options:
                self.declare_partials(of=self._state_output_names[state_name],
                                      wrt=f'controls:{control_name_wrt}')

    def _setup_parameters(self):
        if self._standalone_mode:
            self._configure_parameters()

    def _configure_parameters(self):
        """
        Components do not have configure methods, but since we rely on configure-time introspection to determine
        properties of the states, times, controls, parameters, and timeseries, we need to call this method at
        configure time in the parent ExplicitShooting transcription object.
        """
        # The indices of each parameter in p
        self.p_size = 0
        self.parameter_idxs = {}
        self._parameter_idxs_in_theta = {}
        self._parameter_idxs_in_z = {}
        self._param_input_names = {}

        for param_name, options in self.parameter_options.items():
            self._param_input_names[param_name] = f'parameters:{param_name}'
            self._totals_wrt_names.append(self._param_input_names[param_name])

            self.add_input(self._param_input_names[param_name],
                           shape=options['shape'],
                           val=options['val'],
                           units=options['units'],
                           desc=f'value for parameter {param_name}')

            param_size = np.prod(options['shape'], dtype=int)
            self.parameter_idxs[param_name] = np.s_[self.p_size:self.p_size + param_size]
            self.p_size += param_size

    def _setup_controls(self):
        if self._standalone_mode:
            self._configure_controls()

    def _configure_controls(self):
        """
        Components do not have configure methods, but since we rely on configure-time introspection to determine
        properties of the states, times, controls, parameters, and timeseries, we need to call this method at
        configure time in the parent ExplicitShooting transcription object.
        """
        self.u_size = 0
        self._control_idxs_in_theta = {}
        self._control_idxs_in_z = {}
        self._control_input_names = {}

        for control_name, options in self.control_options.items():
            if options['control_type'] == 'polynomial':
                num_input_nodes = options['order'] + 1
                control_param_shape = (num_input_nodes,) + options['shape']
            else:
                control_param_shape = (self._num_control_input_nodes,) + options['shape']
            control_param_size = np.prod(control_param_shape, dtype=int)
            self._control_input_names[control_name] = f'controls:{control_name}'

            self._totals_wrt_names.append(self._control_input_names[control_name])

            self.add_input(self._control_input_names[control_name],
                           shape=control_param_shape,
                           units=options['units'],
                           desc=f'values for control {control_name} at input nodes')

            self.u_size += control_param_size

    def _build_dx_dz_idxs(self):
        self._partial_dx_dz_idxs = {}

        dx_dz_idx = 0
        for output_state, output_state_options in self.state_options.items():
            output_name = self._state_output_names[output_state]
            output_state_size = np.prod(output_state_options['shape'])

            # Column indices wrt each initial state value
            for input_state, input_state_options in self.state_options.items():
                input_name = self._state_input_names[input_state]
                input_size = np.prod(input_state_options['shape'])
                idxs = np.s_[:, dx_dz_idx: dx_dz_idx + output_state_size * input_size]
                self._partial_dx_dz_idxs[output_name, input_name] = idxs
                dx_dz_idx += output_state_size * input_size

            # Column indices wrt t_initial and t_duration
            self._partial_dx_dz_idxs[output_name, 't_initial'] = np.s_[:, dx_dz_idx]
            self._partial_dx_dz_idxs[output_name, 't_duration'] = np.s_[:, dx_dz_idx + 1]
            dx_dz_idx += 2

            # Column indices wrt the parameters
            for param, param_options in self.parameter_options.items():
                input_name = self._param_input_names[param]
                input_size = np.prod(param_options['shape'])
                idxs = np.s_[:, dx_dz_idx: dx_dz_idx + output_state_size * input_size]
                self._partial_dx_dz_idxs[output_name, input_name] = idxs
                dx_dz_idx += output_state_size * input_size

            # Column indices wrt the controls
            for control, control_options in self.control_options.items():
                if control_options['control_type'] == 'polynomial':
                    input_name = self._control_input_names[control]
                    input_size = np.prod(control_options['shape']) * (control_options['order'] + 1)
                    idxs = np.s_[:, dx_dz_idx: dx_dz_idx + output_state_size * input_size]
                    self._partial_dx_dz_idxs[output_name, input_name] = idxs
                    dx_dz_idx += output_state_size * input_size
                else:
                    input_name = self._control_input_names[control]
                    input_size = np.prod(control_options['shape']) * self._input_grid_data.subset_num_nodes['control_input']
                    idxs = np.s_[:, dx_dz_idx: dx_dz_idx + output_state_size * input_size]
                    self._partial_dx_dz_idxs[output_name, input_name] = idxs
                    dx_dz_idx += output_state_size * input_size

    def _setup_storage(self):
        if self._standalone_mode:
            self._configure_storage()

    def _configure_storage(self):
        igd = self._input_grid_data
        ogd = self._output_grid_data
        control_input_node_ptau = igd.node_ptau[igd.subset_node_indices['control_input']]

        # allocate the ODE parameter vector
        self.theta_size = 2 + self.p_size + self.u_size + self.up_size

        # allocate the integration parameter vector
        self.z_size = self.x_size + self.theta_size

        start_z = 0

        x_output_names = []
        z_input_names = []

        for state_name, options in self.state_options.items():
            state_size = np.prod(options['shape'], dtype=int)
            self._state_idxs_in_z[state_name] = np.s_[start_z: start_z + state_size]
            x_output_names.extend([self._state_output_names[state_name]] * state_size)
            z_input_names.extend([self._state_input_names[state_name]] * state_size)
            start_z += state_size

        # Add 2 to account for t_initial, t_duration
        start_z = self.x_size + 2
        start_theta = 2
        z_input_names.extend(['t_initial', 't_duration'])

        for param_name, options in self.parameter_options.items():
            param_size = np.prod(options['shape'], dtype=int)
            self._parameter_idxs_in_z[param_name] = np.s_[start_z: start_z + param_size]
            self._parameter_idxs_in_theta[param_name] = np.s_[start_theta: start_theta + param_size]
            z_input_names.extend([self._param_input_names[param_name]] * param_size)
            start_z += param_size
            start_theta += param_size

        for control_name, options in self.control_options.items():
            if options['control_type'] == 'polynomial':
                num_input_nodes = options['order'] + 1
                control_param_shape = (num_input_nodes,) + options['shape']
                control_param_size = np.prod(control_param_shape, dtype=int)
                self._control_idxs_in_z[control_name] = np.s_[start_z:start_z + control_param_size]
                self._control_idxs_in_theta[control_name] = np.s_[start_theta:start_theta + control_param_size]
                z_input_names.extend([self._control_input_names[control_name]] * control_param_size)
                start_z += control_param_size
                start_theta += control_param_size
            else:
                control_param_shape = (len(control_input_node_ptau),) + options['shape']
                control_param_size = np.prod(control_param_shape, dtype=int)
                self._control_idxs_in_z[control_name] = np.s_[start_z:start_z + control_param_size]
                self._control_idxs_in_theta[control_name] = np.s_[start_theta:start_theta + control_param_size]
                z_input_names.extend([self._control_input_names[control_name]] * control_param_size)
                start_z += control_param_size
                start_theta += control_param_size

        # Allocate caches to store integrated quantities so we don't have to integrate the same inputs twice
        # if partials are requested for the same inputs as a compute.

        self._nnps = ogd.subset_num_nodes_per_segment['all']

        nn = sum(self._nnps)
        self._t_out = np.zeros((nn, 1))
        self._x_out = np.zeros((nn, self.x_size))

        if self.options['propagate_derivs']:
            self._dx_dz_out = np.zeros((nn, self.x_size * self.z_size))
            self._dt_dz_out = np.zeros((nn, self.z_size))
        else:
            self._dx_dz_out = None
            self._dt_dz_out = None

        # Build a map for obtaining the partials from dx_dz
        self._build_dx_dz_idxs()

        # Allocate the initial values for dx_dz and dt_dz
        self._dx_dz_0 = np.zeros((self.x_size, self.z_size))
        self._dx_dz_0[:, :self.x_size] = np.eye(self.x_size)

        self._dt_dz_0 = np.zeros((1, self.z_size))
        self._dt_dz_0[0, self.x_size] = 1.

        self._dtheta_dz = np.zeros((self.theta_size, self.z_size))
        self._dtheta_dz[:, -self.theta_size:] = np.eye(self.theta_size)

    def setup(self):
        """
        Add the necessary I/O and storage for the ODEIntegrationComp.
        """
        ogd = self._output_grid_data

        # The segment distribution needs to be the same in from the input grid to the output grid.
        if not self._input_grid_data.is_aligned_with(self._output_grid_data):
            raise RuntimeError(f'{self.pathname}: The input grid and the output grid must have the same number of '
                               f'segments and segment spacing, but the input grid segment ends are '
                               f'\n{self._input_grid_data.segment_ends}\n and the output grid segment ends are \n'
                               f'{self._output_grid_data.segment_ends}.')

        self._num_output_rows = ogd.subset_num_nodes['all']

        self._totals_of_names = []
        self._totals_wrt_names = []

        self._setup_subprob()
        self._setup_time()
        self._setup_parameters()
        self._setup_controls()
        self._setup_states()
        self._setup_storage()

    def _subprob_run_model(self, x, t, theta, linearize=True):
        """
        Set inputs to the model given x, t, and theta, evaluate the model, and linearize if requested.

        Parameters
        ----------
        x : np.ndarray
            A flattened, contiguous vector of the state values.
        t : float
            The current time of the integration.
        theta : np.ndarray
            A flattened, contiguous vector of the ODE parameter values.
        linearize : bool
            If True, linearize the model after calling run_model.

        Returns
        -------

        """
        subprob = self._eval_subprob
        t_units = self.time_options['units']
        t_name = self.time_options['name']

        # transcribe time
        subprob.set_val(t_name, t, units=t_units)
        subprob.set_val('t_initial', theta[0], units=t_units)
        subprob.set_val('t_duration', theta[1], units=t_units)

        # transcribe states
        for name in self.state_options:
            input_name = self._state_input_names[name]
            subprob.set_val(input_name, x[0, self.state_idxs[name]])

        # transcribe parameters
        for name in self.parameter_options:
            input_name = self._param_input_names[name]
            subprob.set_val(input_name, theta[self._parameter_idxs_in_theta[name]])

        # transcribe controls
        for name in self.control_options:
            input_name = self._control_input_names[name]
            subprob.set_val(input_name, theta[self._control_idxs_in_theta[name]])

        # Re-run in case the inputs have changed.
        subprob.run_model()

        if linearize:
            subprob.model._linearize(None)

    def eval_ode(self, x, t, theta, eval_solution=True, eval_derivs=True):
        """
        Evaluate the derivative of the ODE output rates wrt the inputs.

        Note that the control parameterization `u` undergoes an interpolation to provide the
        control values at any given time.  The ODE is then a function of these interpolated control
        values, we'll call them `u_hat`.  Technically, the derivatives wrt to `u` need to be chained
        together, but in this implementation the interpolation is part of the execution of the ODE
        and the chained derivatives are captured correctly there.

        Parameters
        ----------
        x : np.ndarray
            A flattened, contiguous vector of the state values.
        t : float
            The current time of the integration.
        theta : np.ndarray
            A flattened, contiguous vector of the ODE parameter values.
        eval_solution : bool
            If True, return the state rate output values, otherwise the first output will be None.
        eval_derivs : bool
            If True, return the derivatives of the state rate outputs wrt x, t, and theta. Otherwise the final
            three outputs will be None.

        Returns
        -------
        f : np.ndarray
            The outputs of the ODE state rates, or None if eval_solution is False.
        f_x : np.ndarray
            A matrix of the derivative of each element of the rates `f` wrt each value in `x`, or None
            if eval_derivs is False.
        f_t : np.ndarray
            A matrix of the derivatives of each element of the rates `f` wrt `time`, or None
            if eval_derivs is False.
        f_theta : np.ndarray
            A matrix of the derivatives of each element of the rates `f` wrt the parameters `theta`, or None
            if eval_derivs is False.
        """
        t_name = self.time_options['name']
        self._subprob_run_model(x, t, theta, linearize=False)

        # pack the resulting array
        if eval_solution:
            f = np.zeros((self.x_size, 1))
            for name in self.state_options:
                state_rate = self._eval_subprob.get_val(f'state_rate_collector.state_rates:{name}_rate')
                f[self.state_idxs[name]] = state_rate.reshape((self.state_sizes[name], 1))
        else:
            f = None

        if eval_derivs:
            f_t = np.zeros((self.x_size, 1))
            f_x = np.zeros((self.x_size, self.x_size))
            f_theta = np.zeros((self.x_size, self.theta_size))

            totals = self._eval_subprob.compute_totals(of=self._state_rate_of_names,
                                                       wrt=self._totals_wrt_names,
                                                       use_abs_names=False)

            for state_name in self.state_options:
                of_name = f'state_rate_collector.state_rates:{state_name}_rate'
                idxs = self.state_idxs[state_name]

                f_t[self.state_idxs[state_name]] = totals[of_name, t_name]

                for state_name_wrt in self.state_options:
                    idxs_wrt = self.state_idxs[state_name_wrt]
                    px_px = totals[of_name, self._state_input_names[state_name_wrt]]
                    f_x[idxs, idxs_wrt] = px_px.ravel()

                f_theta[idxs, 0] = totals[of_name, 't_initial']
                f_theta[idxs, 1] = totals[of_name, 't_duration']

                for param_name_wrt in self.parameter_options:
                    idxs_wrt = self._parameter_idxs_in_theta[param_name_wrt]
                    px_pp = totals[of_name, self._param_input_names[param_name_wrt]]
                    f_theta[idxs, idxs_wrt] = px_pp.ravel()

                for control_name_wrt in self.control_options:
                    idxs_wrt = self._control_idxs_in_theta[control_name_wrt]
                    px_pu = totals[of_name, self._control_input_names[control_name_wrt]]
                    f_theta[idxs, idxs_wrt] = px_pu.ravel()

        else:
            f_x = f_t = f_theta = None

        return f, f_x, f_t, f_theta

    def _f_augmented(self, t, y, theta, dtheta_dz):
        """
        The ODE-callable function where y is the augmented state vector, theta are the ODE parameters, and dtheta_dz
        are the sensitivities of the ODE parameters to the integration parameters.

        Parameters
        ----------
        t : float
            The current value of the integration variable.
        y : np.array
            The augmented state vector.
        theta : np.array
            The ODE parameter vector. The first two elements are t_initial and t_duration.
        dtheta_dz : np.array
            The sensitivities of the ODE parameters wrt the integration parameters. Since the ODE parameters are
            just the integration parameters with the first num_x columns removed, this is a matrix of shape
            (num_dtheta, num_z - num_x).

        Returns
        -------
        y_dot : np.array
            The rates associated with each state in the augmented state vector (primal and tangent states).
        """
        n_x = self.x_size
        n_theta = self.theta_size
        n_z = n_x + n_theta

        x = y[:n_x].reshape((1, n_x))
        td = theta[1]

        dx_dz = y[n_x:n_x + n_x * n_z].reshape((n_x, n_z))
        dt_dz = y[-n_z:].reshape((1, n_z))

        x_dot, f_x, f_t, f_theta = self.eval_ode(x, t, theta, eval_solution=True, eval_derivs=True)

        dh_dz = np.zeros((1, n_z))
        dh_dz[0, n_x + 1] = 1. / td
        dt_dz_dot = dh_dz

        dx_dz_dot = f_x @ dx_dz + f_t @ dt_dz + f_theta @ dtheta_dz + x_dot @ dt_dz_dot

        y_dot = np.concatenate((x_dot.ravel(),
                                dx_dz_dot.ravel(),
                                dt_dz_dot.ravel()))

        return y_dot

    def _f_primal(self, t, x, theta):
        """
        The ODE-callable function where y is the augmented state vector, theta are the ODE parameters, and dtheta_dz
        are the sensitivities of the ODE parameters to the integration parameters.

        Parameters
        ----------
        t : float
            The current value of the integration variable.
        x : np.array
            The primal state vector.
        theta : np.array
            The ODE parameter vector. The first two elements are t_initial and t_duration.

        Returns
        -------
        y_dot : np.array
            The rates associated with each state in the augmented state vector (primal and tangent states).
        """
        n_x = self.x_size
        _x = x.reshape((1, n_x))

        x_dot, _, _, _ = self.eval_ode(_x, t, theta, eval_solution=True, eval_derivs=False)

        return x_dot.ravel()

    def _propagate(self, inputs, propagate_derivs=None, x_out=None, t_out=None, dx_dz_out=None,
                   dt_dz_out=None):
        """
        Propagate the augmented state, y = [x, dx_dz, dt_dz].

        Parameters
        ----------
        inputs : Vector
            The inputs of the integration. This is an OpenMDAO vector including the initial state values, t_initial
            and t_duration, the ODE parameters, the control values, and the polynomial control values.
        propagate_derivs : bool or None
            If True, propagate the derivatives of the states and of time w.r.t. the integration parameters vector, z.
            If None, use the value of option `propagate_derivs` to determine if derivatives should be computed.
        x_out : np.array or None.
            Pre-allocated storage for the results of the state propagation. If None, this storage will be allocated
            on each call to _propagate.
        t_out : np.array or None.
            Pre-allocated storage for the time at the requested nodes. If None, this storage will be allocated
            on each call to _propagate.
        dx_dz_out : np.array or None.
            Pre-allocated storage for the results of the propagation of derivatives of the states with respect to the
            integration parameters. If None, this storage will be allocated on each call to _propagate.
        dt_dz_out : np.array or None.
            Pre-allocated storage derivatives of the time at the requested nodes with respect to the integration
            parameters. If None, this storage will be allocated on each call to _propagate.

        Returns
        -------
        x_out : np.array
            The integrated states at each node.
        t_out : np.array
            The time (or integration variable) value at each node.
        dx_dz_out : np.array
            The derivative of each state at each node with respect to the initial integration parameters.
        dt_dz_out : np.array
            The derivative of time at each node with respect to the initial integration parameters.
        """
        method = self.options['method']
        first_step = self.options['first_step']
        max_step = self.options['max_step']
        atol = self.options['atol']
        rtol = self.options['rtol']
        ogd = self._output_grid_data

        nn = sum(self._nnps)
        if x_out is None:
            x_out = np.zeros((nn, self.x_size))
        if t_out is None:
            t_out = np.zeros((nn, 1))

        if propagate_derivs is None:
            _propagate_derivs = self.options['propagate_derivs']
        else:
            _propagate_derivs = propagate_derivs

        n_x = self.x_size
        n_theta = self.theta_size
        n_z = n_x + n_theta
        nnps = self._nnps

        # Extract the input values
        x0 = np.zeros((self.x_size,))
        theta = np.zeros((self.theta_size,))

        for state_name in self.state_options:
            state_initial_val = inputs[self._state_input_names[state_name]]
            x0[self.state_idxs[state_name]] = state_initial_val.ravel()

        theta[0] = t_initial = inputs['t_initial'].ravel()[0]
        theta[1] = t_duration = inputs['t_duration'].ravel()[0]

        for param_name in self.parameter_options:
            param_val = inputs[self._param_input_names[param_name]]
            theta[self._parameter_idxs_in_theta[param_name]] = param_val.ravel()

        for control_name in self.control_options:
            control_vals = inputs[self._control_input_names[control_name]]
            theta[self._control_idxs_in_theta[control_name]] = control_vals.ravel()

        if _propagate_derivs:
            if dx_dz_out is None:
                dx_dz_out = np.zeros((nn, self.x_size * self.z_size))
            if dt_dz_out is None:
                dt_dz_out = np.zeros((nn, self.z_size))

            y0 = np.concatenate((x0.ravel(), self._dx_dz_0.ravel(), self._dt_dz_0.ravel()))
        else:
            dx_dz_out = None
            dt_dz_out = None
            y0 = x0.ravel()

        row_seg_i = 0

        for i in range(ogd.num_segments):
            self._set_segment_index(i)

            eval_nodes_ptau = ogd.node_ptau[ogd.segment_indices[i, 0]: ogd.segment_indices[i, 1]]

            t_eval_seg = t_initial + 0.5 * (eval_nodes_ptau + 1) * t_duration
            t_span_seg = (t_eval_seg[0], t_eval_seg[-1])

            if _propagate_derivs:
                # The augmented initial state vector
                sol = solve_ivp(self._f_augmented, t_span=t_span_seg, t_eval=t_eval_seg, y0=y0,
                                args=(theta, self._dtheta_dz), method=method, first_step=first_step, max_step=max_step,
                                atol=atol, rtol=rtol)
                dx_dz_out[row_seg_i:row_seg_i + nnps[i], :] = sol.y.T[:, n_x:n_x + n_x * n_z]
                dt_dz_out[row_seg_i:row_seg_i + nnps[i], :] = sol.y.T[:, -n_z:]
            else:
                sol = solve_ivp(self._f_primal, t_span=t_span_seg, t_eval=t_eval_seg, y0=y0, args=(theta,),
                                method=method, first_step=first_step, max_step=max_step, atol=atol, rtol=rtol)

            if not sol.success:
                raise om.AnalysisError(f'solve_ivp failed: {sol.message}')

            x_out[row_seg_i:row_seg_i + nnps[i], :] = sol.y.T[:, :n_x]  # Save solution to the output nodes
            t_out[row_seg_i:row_seg_i + nnps[i], 0] = sol.t
            y0 = sol.y.T[-1, :]  # Set initial y for the next segment
            row_seg_i += nnps[i]  # Increment node associated with the start of the next segment

        return x_out, t_out, dx_dz_out, dt_dz_out

    def compute(self, inputs, outputs):
        """
        Compute propagated state values.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        t_name = self.time_options['name']

        inputs_hash = inputs.get_hash()
        if inputs_hash != self._inputs_cache:
            self._propagate(inputs=inputs, propagate_derivs=self.options['propagate_derivs'],
                            x_out=self._x_out, t_out=self._t_out, dx_dz_out=self._dx_dz_out, dt_dz_out=self._dt_dz_out)

            self._inputs_cache = inputs_hash

        t = self._t_out
        x = self._x_out

        # Extract time
        outputs[t_name] = t
        outputs[f'{t_name}_phase'] = t - inputs['t_initial']
        outputs['t_final'] = inputs['t_initial'] + inputs['t_duration']

        # Extract the state values
        for state_name in self.state_options:
            of = self._state_output_names[state_name]
            outputs[of] = x[:, self.state_idxs[state_name]]

    def compute_partials(self, inputs, partials):
        """
        Compute derivatives of propagated states wrt the inputs.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        """
        # Only propagate the ODE if our inputs have changed, otherwise use the cached outputs.
        inputs_hash = inputs.get_hash()
        if inputs_hash != self._inputs_cache:
            self._propagate(inputs=inputs, propagate_derivs=True,
                            x_out=self._x_out, t_out=self._t_out, dx_dz_out=self._dx_dz_out, dt_dz_out=self._dt_dz_out)
            self._inputs_cache = inputs_hash

        t_name = self.time_options['name']

        dt_dz = self._dt_dz_out
        dx_dz = self._dx_dz_out

        partials[t_name, 't_duration'] = dt_dz[:, self.x_size + 1]
        partials[f'{t_name}_phase', 't_duration'] = dt_dz[:, self.x_size + 1]

        for state_name in self.state_options:
            of = self._state_output_names[state_name]

            for wrt_state_name in self.state_options:
                wrt = self._state_input_names[wrt_state_name]
                partials[of, wrt] = dx_dz[self._partial_dx_dz_idxs[of, wrt]]

            partials[of, 't_initial'] = dx_dz[self._partial_dx_dz_idxs[of, 't_initial']]
            partials[of, 't_duration'] = dx_dz[self._partial_dx_dz_idxs[of, 't_duration']]

            for wrt_param_name in self.parameter_options:
                wrt = self._param_input_names[wrt_param_name]
                partials[of, wrt] = dx_dz[self._partial_dx_dz_idxs[of, wrt]]

            for wrt_control_name in self.control_options:
                wrt = self._control_input_names[wrt_control_name]
                partials[of, wrt] = dx_dz[self._partial_dx_dz_idxs[of, wrt]]



import openmdao.api as om
from .mdot_comp import MassFlowRateComp
from .max_thrust_comp import MaxThrustComp
from .thrust_comp import ThrustComp


class PropGroup(om.Group):
    """
    The purpose of the PropGroup is to compute the propulsive forces on the
    aircraft in the body frame.

    Parameters
    ----------
    mach : float
        Mach number (unitless)
    alt : float
        altitude (m)
    Isp : float
        specific impulse (s)
    throttle : float
        throttle value nominally between 0.0 and 1.0 (unitless)

    Unknowns
    --------
    thrust : float
        Vehicle thrust force (N)
    mdot : float
        Vehicle mass accumulation rate (kg/s)

    """
    def initialize(self):
        self.options.declare('num_nodes', types=int,
                             desc='Number of nodes to be evaluated in the RHS')

    def setup(self):
        nn = self.options['num_nodes']

        max_thrust_comp = MaxThrustComp(vec_size=nn, extrapolate=True, method='cubic')
        # max_thrust_comp = BrysonMaxThrustComp(num_nodes=nn)

        self.add_subsystem(name='max_thrust_comp',
                           subsys=max_thrust_comp,
                           promotes_inputs=['mach', 'h'],
                           promotes_outputs=['max_thrust'])

        self.add_subsystem(name='thrust_comp',
                           subsys=ThrustComp(num_nodes=nn),
                           promotes_inputs=['max_thrust', 'throttle'],
                           promotes_outputs=['thrust'])

        self.add_subsystem(name='mdot_comp',
                           subsys=MassFlowRateComp(num_nodes=nn),
                           promotes_inputs=['thrust', 'Isp'],
                           promotes_outputs=['m_dot'])



import unittest

try:
    import matplotlib
except ImportError:
    matplotlib = None

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.assert_utils import assert_near_equal


@use_tempdirs
class TestRaceCarForDocs(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_racecar_for_docs(self):
        import openmdao.api as om
        import dymos as dm

        from dymos.examples.racecar.combinedODE import CombinedODE
        from dymos.examples.racecar.spline import get_spline, get_track_points
        from dymos.examples.racecar.tracks import ovaltrack  # track curvature imports

        # change track here and in curvature.py. Tracks are defined in tracks.py
        track = ovaltrack

        # generate nodes along the centerline for curvature calculation (different
        # than collocation nodes)
        points = get_track_points(track)

        # fit the centerline spline.
        finespline, gates, gatesd, curv, slope = get_spline(points, s=0.0)
        # by default 10000 points
        s_final = track.get_total_length()

        txs = {'radau': dm.Radau(num_segments=50, order=3, compressed=True),
               'gauss-lobatto': dm.GaussLobatto(num_segments=50, order=3, compressed=True)}

        for tx_name, tx in txs.items():
            with self.subTest(tx_name):
                # Define the OpenMDAO problem
                p = om.Problem(model=om.Group())

                # Define a Trajectory object
                traj = dm.Trajectory()
                p.model.add_subsystem('traj', subsys=traj)

                phase = dm.Phase(ode_class=CombinedODE,
                                 transcription=tx)

                traj.add_phase(name='phase0', phase=phase)

                # Set the time options, in this problem we perform a change of variables. So 'time' is
                # actually 's' (distance along the track centerline)
                # This is done to fix the collocation nodes in space, which saves us the calculation of
                # the rate of change of curvature.
                # The state equations are written with respect to time, the variable change occurs in
                # timeODE.py
                phase.set_integ_var_options(fix_initial=True, fix_duration=True, duration_val=s_final,
                                            name='s', targets=['curv.s'], units='m', duration_ref=s_final,
                                            duration_ref0=10)

                # Define states
                phase.add_state('t', fix_initial=True, fix_final=False, units='s', lower=0,
                                rate_source='dt_ds', ref=100)  # time
                phase.add_state('n', fix_initial=False, fix_final=False, units='m', upper=4.0, lower=-4.0,
                                rate_source='dn_ds', targets=['n'],
                                ref=4.0)  # normal distance to centerline. The bounds on n define the
                # width of the track
                phase.add_state('V', fix_initial=False, fix_final=False, units='m/s', ref=40, ref0=5,
                                rate_source='dV_ds', targets=['V'])  # velocity
                phase.add_state('alpha', fix_initial=False, fix_final=False, units='rad',
                                rate_source='dalpha_ds', targets=['alpha'],
                                ref=0.15)  # vehicle heading angle with respect to centerline
                phase.add_state('lambda', fix_initial=False, fix_final=False, units='rad',
                                rate_source='dlambda_ds', targets=['lambda'],
                                ref=0.01)  # vehicle slip angle, or angle between the axis of the vehicle
                # and velocity vector (all cars drift a little)
                phase.add_state('omega', fix_initial=False, fix_final=False, units='rad/s',
                                rate_source='domega_ds', targets=['omega'], ref=0.3)  # yaw rate
                phase.add_state('ax', fix_initial=False, fix_final=False, units='m/s**2',
                                rate_source='dax_ds', targets=['ax'], ref=8)  # longitudinal acceleration
                phase.add_state('ay', fix_initial=False, fix_final=False, units='m/s**2',
                                rate_source='day_ds', targets=['ay'], ref=8)  # lateral acceleration

                # Define Controls
                phase.add_control(name='delta', units='rad', lower=None, upper=None, fix_initial=False,
                                  fix_final=False, ref=0.04, rate_continuity=True)  # steering angle
                phase.add_control(name='thrust', units=None, fix_initial=False, fix_final=False, rate_continuity=True)
                # the thrust controls the longitudinal force of the rear tires and
                # is positive while accelerating, negative while braking

                # Performance Constraints
                pmax = 960000  # W
                phase.add_path_constraint('power', upper=pmax, ref=100000)  # engine power limit

                # The following four constraints are the tire friction limits, with 'rr' designating the
                # rear right wheel etc. This limit is computed in tireConstraintODE.py
                phase.add_path_constraint('c_rr', upper=1)
                phase.add_path_constraint('c_rl', upper=1)
                phase.add_path_constraint('c_fr', upper=1)
                phase.add_path_constraint('c_fl', upper=1)

                # Some of the vehicle design parameters are available to set here. Other parameters can
                # be found in their respective ODE files.
                phase.add_parameter('M', val=800.0, units='kg', opt=False,
                                    targets=['car.M', 'tire.M', 'tireconstraint.M', 'normal.M'],
                                    static_target=True)  # vehicle mass
                phase.add_parameter('beta', val=0.62, units=None, opt=False, targets=['tire.beta'],
                                    static_target=True)  # brake bias
                phase.add_parameter('CoP', val=1.6, units='m', opt=False, targets=['normal.CoP'],
                                    static_target=True)  # center of pressure location
                phase.add_parameter('h', val=0.3, units='m', opt=False, targets=['normal.h'],
                                    static_target=True)  # center of gravity height
                phase.add_parameter('chi', val=0.5, units=None, opt=False, targets=['normal.chi'],
                                    static_target=True)  # roll stiffness
                phase.add_parameter('ClA', val=4.0, units='m**2', opt=False, targets=['normal.ClA'],
                                    static_target=True)  # downforce coefficient*area
                phase.add_parameter('CdA', val=2.0, units='m**2', opt=False, targets=['car.CdA'],
                                    static_target=True)  # drag coefficient*area

                # Minimize final time.
                # note that we use the 'state' time instead of Dymos 'time'
                phase.add_objective('t', loc='final')

                # Add output timeseries
                phase.add_timeseries_output('t', output_name='time')

                # Link the states at the start and end of the phase in order to ensure a continous lap
                traj.link_phases(phases=['phase0', 'phase0'],
                                 vars=['V', 'n', 'alpha', 'omega', 'lambda', 'ax', 'ay'],
                                 locs=('final', 'initial'))

                # Set the driver. IPOPT or SNOPT are recommended but SLSQP might work.
                p.driver = om.pyOptSparseDriver(optimizer='IPOPT')

                p.driver.opt_settings['mu_init'] = 1e-3
                p.driver.opt_settings['max_iter'] = 500
                p.driver.opt_settings['acceptable_tol'] = 1e-3
                p.driver.opt_settings['constr_viol_tol'] = 1e-3
                p.driver.opt_settings['compl_inf_tol'] = 1e-3
                p.driver.opt_settings['acceptable_iter'] = 0
                p.driver.opt_settings['tol'] = 1e-3
                p.driver.opt_settings['print_level'] = 0
                p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'  # for faster convergence
                p.driver.opt_settings['alpha_for_y'] = 'safer-min-dual-infeas'
                p.driver.opt_settings['mu_strategy'] = 'monotone'
                p.driver.opt_settings['bound_mult_init_method'] = 'mu-based'
                # p.driver.options['print_results'] = False

                # Allow OpenMDAO to automatically determine our sparsity pattern.
                # Doing so can significant speed up the execution of Dymos.
                p.driver.declare_coloring()

                # Setup the problem
                p.setup(check=True)  # force_alloc_complex=True
                # Now that the OpenMDAO problem is setup, we can set the values of the states.

                # Test that set_integ_var_val works.
                # This isn't necessary because we already set the fixed endpoints
                # in set_integ_var_options, but here it's serving as a test of the method.
                phase.set_integ_var_val(initial=0, duration=s_final, units='m')

                # States
                # Nonzero velocity to avoid division by zero errors
                phase.set_state_val('V', 20.0, units='m/s')
                # All other states start at 0
                phase.set_state_val('lambda', 0.0, units='rad')
                phase.set_state_val('omega', 0.0, units='rad/s')
                phase.set_state_val('alpha', 0.0, units='rad')
                phase.set_state_val('ax', 0.0, units='m/s**2')
                phase.set_state_val('ay', 0.0, units='m/s**2')
                phase.set_state_val('n', 0.0, units='m')
                # initial guess for what the final time should be
                phase.set_state_val('t', [0.0, 100.0], units='s')

                # Controls
                # a small amount of thrust can speed up convergence
                phase.set_control_val('delta', 0.0, units='rad')
                phase.set_control_val('thrust', 0.1, units=None)

                dm.run_problem(p, run_driver=True)

                assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1, ...], 22.2657, tolerance=1.0E-2)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
import openmdao.api as om

from .birkhoff_defect_comp import BirkhoffDefectComp

from ...grid_data import GridData
from ....phase.options import TimeOptionsDictionary
from ....utils.ode_utils import _make_ode_system


class BirkhoffIterGroup(om.Group):
    """
    Class definition for the BirkhoffIterGroup.

    This group allows for iteration of the state variables and initial _or_ final value of the state
    depending on the direction of the solve.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._implicit_outputs = set()

    def initialize(self):
        """
        Declare group options.
        """
        self.options.declare('state_options', types=dict,
                             desc='Dictionary of options for the states.')
        self.options.declare('time_options', types=TimeOptionsDictionary,
                             desc='Options for time in the phase.')
        self.options.declare('parameter_options', types=dict, default={},
                             desc='parameter options inherited from the phase')
        self.options.declare('grid_data', types=GridData, desc='Container object for grid info.')
        self.options.declare('ode_class', default=None,
                             desc='Callable that instantiates the ODE system.',
                             recordable=False)
        self.options.declare('ode_init_kwargs', types=dict, default={},
                             desc='Keyword arguments provided when initializing the ODE System')
        self.options.declare('calc_exprs', types=dict, default={},
                             desc='ODE Expresions from the Phase')

    def setup(self):
        """
        Define the structure of the control group.
        """
        gd = self.options['grid_data']
        nn = gd.subset_num_nodes['all']
        state_options = self.options['state_options']
        time_options = self.options['time_options']
        ode_class = self.options['ode_class']
        ode_init_kwargs = self.options['ode_init_kwargs']

        ode = _make_ode_system(ode_class=ode_class,
                               num_nodes=nn,
                               ode_init_kwargs=ode_init_kwargs,
                               calc_exprs=self.options['calc_exprs'],
                               parameter_options=self.options['parameter_options'])

        self.add_subsystem('ode_all', subsys=ode)

        self.add_subsystem('collocation_comp',
                           subsys=BirkhoffDefectComp(grid_data=gd,
                                                     state_options=state_options,
                                                     time_units=time_options['units']),
                           promotes_inputs=['*'], promotes_outputs=['*'])

        if any([opts['solve_segments'] in ('forward', 'backward') for opts in state_options.values()]):
            self.add_subsystem('states_balance_comp', subsys=om.InputResidsComp(),
                               promotes_inputs=['*'], promotes_outputs=['*'])

    def _configure_desvars(self, name, options):
        state_name = f'states:{name}'
        initial_state_name = f'initial_states:{name}'
        final_state_name = f'final_states:{name}'
        state_rate_name = f'state_rates:{name}'

        solve_segs = options['solve_segments']
        opt = options['opt']

        num_nodes = self.options['grid_data'].subset_num_nodes['col']

        ib = (None, None) if options['initial_bounds'] is None else options['initial_bounds']
        fb = (None, None) if options['final_bounds'] is None else options['final_bounds']
        lower = options['lower']
        upper = options['upper']
        scaler = options['scaler']
        adder = options['adder']
        ref0 = options['ref0']
        ref = options['ref']
        fix_initial = options['fix_initial']
        fix_final = options['fix_final']
        input_initial = options['input_initial']
        input_final = options['input_final']
        shape = options['shape']

        if solve_segs == 'forward' and fix_final:
            raise ValueError(f"Option fix_final on state {name} may not "
                             f"be used with `solve_segments='forward'`.\n Use "
                             f"a boundary constraint to constrain the final "
                             f"state value instead.")
        elif solve_segs == 'backward' and fix_initial:
            raise ValueError(f"Option fix_final on state {name} may not "
                             f"be used with `solve_segments='forward'`.\n Use "
                             f"a boundary constraint to constrain the initial "
                             f"state value instead.")

        if not np.isscalar(ref0) and ref0 is not None:
            ref0 = np.asarray(ref0)
            if ref0.shape == shape:
                ref0_state = np.tile(ref0.flatten(), num_nodes)
                # ref0_seg_ends = np.tile(ref0.flatten(), 2)
            else:
                raise ValueError('array-valued scaler/ref must length equal to state-size')
        else:
            ref0_state = ref0
            # ref0_seg_ends = ref0
        if not np.isscalar(ref) and ref is not None:
            ref = np.asarray(ref)
            if ref.shape == shape:
                ref_state = np.tile(ref.flatten(), num_nodes)
                # ref_seg_ends = np.tile(ref.flatten(), 2)
            else:
                raise ValueError('array-valued scaler/ref must length equal to state-size')
        else:
            ref_state = ref
            # ref_seg_ends = ref

        free_vars = {state_name, state_rate_name, initial_state_name, final_state_name}

        if solve_segs == 'forward':
            implicit_outputs = {state_name, state_rate_name, final_state_name}
        elif solve_segs == 'backward':
            implicit_outputs = {state_name, state_rate_name, initial_state_name}
        else:
            implicit_outputs = set()

        free_vars = free_vars - implicit_outputs

        if fix_initial or input_initial:
            free_vars = free_vars - {initial_state_name}
        if fix_final or input_final:
            free_vars = free_vars - {final_state_name}

        if opt:
            # Add design variables for the remaining free variables
            if state_name in free_vars:
                self.add_design_var(name=state_name,
                                    lower=lower,
                                    upper=upper,
                                    scaler=scaler,
                                    adder=adder,
                                    ref0=ref0_state,
                                    ref=ref_state)

            if state_rate_name in free_vars:
                self.add_design_var(name=state_rate_name,
                                    scaler=scaler,
                                    adder=adder,
                                    ref0=ref0_state,
                                    ref=ref_state)

            if initial_state_name in free_vars:
                self.add_design_var(name=initial_state_name,
                                    lower=ib[0],
                                    upper=ib[1],
                                    scaler=scaler,
                                    adder=adder,
                                    ref0=ref0,
                                    ref=ref)

            if final_state_name in free_vars:
                self.add_design_var(name=final_state_name,
                                    lower=fb[0],
                                    upper=fb[1],
                                    scaler=scaler,
                                    adder=adder,
                                    ref0=ref0,
                                    ref=ref)

        return implicit_outputs

    def configure_io(self, phase):
        """
        I/O creation is delayed until configure so that we can determine shape and units for the states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        collocation_comp = self._get_subsystem('collocation_comp')
        collocation_comp.configure_io(phase)

        gd = self.options['grid_data']
        nn = gd.subset_num_nodes['all']
        ns = gd.num_segments

        state_options = self.options['state_options']
        states_balance_comp = self._get_subsystem('states_balance_comp')

        for name, options in state_options.items():
            units = options['units']
            rate_source = options['rate_source']
            shape = options['shape']

            for tgt in options['targets']:
                self.promotes('ode_all', [(tgt, f'states:{name}')])
                self.set_input_defaults(f'states:{name}', val=1.0, units=units, src_shape=(nn,) + shape)

            self._implicit_outputs = self._configure_desvars(name, options)

            if f'states:{name}' in self._implicit_outputs:
                states_balance_comp.add_output(f'states:{name}',
                                               shape=(nn,) + shape,
                                               units=units)

                states_balance_comp.add_input(f'state_defects:{name}',
                                              shape=(nn+ns,) + shape,
                                              units=units)

                if ns > 1:
                    states_balance_comp.add_input(f'state_cnty_defects:{name}',
                                                  shape=(ns - 1,) + shape,
                                                  units=units)

            if f'state_rates:{name}' in self._implicit_outputs:
                states_balance_comp.add_output(f'state_rates:{name}', shape=(nn,) + shape, units=units)
                states_balance_comp.add_input(f'state_rate_defects:{name}',
                                              shape=(nn,) + shape,
                                              units=units)

            if f'initial_states:{name}' in self._implicit_outputs:
                states_balance_comp.add_output(f'initial_states:{name}', shape=(1,) + shape, units=units)

            if f'final_states:{name}' in self._implicit_outputs:
                states_balance_comp.add_output(f'final_states:{name}', shape=(1,) + shape, units=units)

            try:
                rate_source_var = options['rate_source']
            except RuntimeError:
                raise ValueError(f"state '{name}' in phase '{phase.name}' was not given a "
                                 "rate_source")

            # Note the rate source must be shape-compatible with the state
            var_type = phase.classify_var(rate_source_var)

            if var_type == 'ode':
                self.connect(f'ode_all.{rate_source}', f'f_computed:{name}')



import unittest

try:
    import matplotlib.pyplot as plt
    plt.switch_backend('Agg')
except ImportError:
    plt = None

from openmdao.utils.testing_utils import use_tempdirs
import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal

import dymos as dm
from dymos.examples.cannonball.size_comp import CannonballSizeComp
from dymos.examples.cannonball.cannonball_ode import CannonballODE


@use_tempdirs
class TestCannonballConnectedInitialStates(unittest.TestCase):

    def test_cannonball_connected_initial_states(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        p.model.add_subsystem('size_comp', CannonballSizeComp(),
                              promotes_inputs=['radius', 'dens'])
        p.model.set_input_defaults('dens', val=7.87, units='g/cm**3')
        p.model.add_design_var('radius', lower=0.01, upper=0.10,
                               ref0=0.01, ref=0.10, units='m')

        traj = dm.Trajectory()
        p.model.add_subsystem('traj', traj)

        transcription = dm.Radau(num_segments=5, order=3, compressed=True)
        ascent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        ascent = traj.add_phase('ascent', ascent)

        ascent.set_time_options(fix_initial=True, duration_bounds=(1, 100),
                                duration_ref=100, units='s')

        ascent.set_time_options(fix_initial=True, duration_bounds=(1, 100), duration_ref=100, units='s')
        ascent.add_state('r', fix_initial=True, fix_final=False, rate_source='r_dot', units='m')
        ascent.add_state('h', fix_initial=True, fix_final=False, units='m', rate_source='h_dot')
        ascent.add_state('gam', fix_initial=False, fix_final=True, units='rad', rate_source='gam_dot')
        ascent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        ascent.add_parameter('S', units='m**2', static_targets=True)
        ascent.add_parameter('m', units='kg', static_targets=True)

        # Limit the muzzle energy
        ascent.add_boundary_constraint('ke2 = 0.5 * m * v**2', loc='initial',
                                       upper=400000, lower=0, ref=100000)

        # Second Phase (descent)
        transcription = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        descent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        traj.add_phase('descent', descent)

        descent.set_time_options(initial_bounds=(.5, 100), duration_bounds=(.5, 100),
                                 duration_ref=100, units='s')
        descent.add_state('r', units='m', rate_source='r_dot', input_initial=True)
        descent.add_state('h', units='m', rate_source='h_dot', input_initial=True, fix_final=True)
        descent.add_state('gam', units='rad', rate_source='gam_dot', input_initial=True, fix_final=False)
        descent.add_state('v', units='m/s', rate_source='v_dot', input_initial=True, fix_final=False)

        descent.add_parameter('S', units='m**2', static_targets=True)
        descent.add_parameter('m', units='kg', static_targets=True)

        descent.add_objective('r', loc='final', scaler=-1.0)

        traj.add_parameter('CD',
                           targets={'ascent': ['CD'], 'descent': ['CD']},
                           val=0.5, units=None, opt=False, static_targets=True)

        traj.add_parameter('m', units='kg', val=1.0,
                           targets={'ascent': 'm', 'descent': 'm'}, static_targets=True)

        traj.add_parameter('S', units='m**2', val=0.005, static_targets=True)

        # Link Phases (link time and all state variables)
        traj.link_phases(phases=['ascent', 'descent'], vars=['time'])
        traj.link_phases(phases=['ascent', 'descent'], vars=['r', 'h', 'v', 'gam'], connected=True)

        p.model.connect('size_comp.mass', 'traj.parameters:m')
        p.model.connect('size_comp.S', 'traj.parameters:S')

        p.model.linear_solver = om.DirectSolver()

        # Finish Problem Setup
        p.setup()

        # Set constants and initial guesses
        p.set_val('radius', 0.05, units='m')
        p.set_val('dens', 7.87, units='g/cm**3')

        traj.set_parameter_val('CD', 0.5)

        ascent.set_time_val(initial=0.0, duration=10.0)

        ascent.set_state_val('r', [0, 100])
        ascent.set_state_val('h', [0, 100])
        ascent.set_state_val('v', [200, 150])
        ascent.set_state_val('gam', [25, 0], units='deg')

        descent.set_time_val(initial=10.0, duration=10.0)

        descent.set_state_val('r', [100, 200])
        descent.set_state_val('h', [100, 0])
        descent.set_state_val('v', [150, 200])
        descent.set_state_val('gam', [0, -45], units='deg')

        # Run the optimization and final explicit simulation
        dm.run_problem(p)

        assert_near_equal(p.get_val('traj.descent.states:r')[-1],
                          3183.25, tolerance=1.0E-2)


if __name__ == '__main__':
    unittest.main()



import numpy as np


def lagrange_matrices(x_disc, x_interp, compute_interp_matrix=True, compute_diff_matrix=True):
    """
    Compute the lagrange matrices.

    The lagrange matrics are given 'discretization' nodes at which
    values are specified and 'interpolation' nodes at which values will be desired,
    returns interpolation and differentiation matrices which provide polynomial
    values and derivatives.

    Parameters
    ----------
    x_disc : np.array
        The cardinal nodes at which values of the variable are specified.
    x_interp : np.array
        The interior nodes at which interpolated values of the variable or its derivative
        are desired.
    compute_interp_matrix : bool
        If True, construct and return the interpolation matrix, otherwise return None.
    compute_diff_matrix : bool
        If True, construct and return the differentiation matrix, otherwise return None. The
        differentiation matrix can be prohibitively expensive to build as the number of
        discretization points grows large.

    Returns
    -------
    np.array
        A num_i x num_c matrix which, when post-multiplied by values specified
        at x_disc, returns the intepolated values at x_interp.

    np.array or None
        A num_i x num_c matrix which, when post-multiplied by values specified
        at x_disc, returns the intepolated derivatives at x_interp. This is returned
        as None if compute_diff_matrix is None.
    """
    nd = len(x_disc)
    ni = len(x_interp)

    if compute_interp_matrix or compute_diff_matrix:
        temp = np.zeros((ni, nd))
        # Barycentric Weights
        diff = np.reshape(x_disc, (nd, 1)) - np.reshape(x_disc, (1, nd))
        np.fill_diagonal(diff, 1.0)
        wb = np.prod(1.0 / diff, axis=1)

    # Compute Li
    diff = np.reshape(x_interp, (ni, 1)) - np.reshape(x_disc, (1, nd))
    if compute_interp_matrix:
        Li = np.zeros((ni, nd))
        for j in range(nd):
            temp[:] = diff[:]
            temp[:, j] = 1.0
            Li[:, j] = wb[j] * np.prod(temp, axis=1)
    else:
        Li = None

    # Compute Di
    if compute_diff_matrix:
        Di = np.zeros((ni, nd))
        for j in range(nd):
            for k in range(nd):
                if k != j:
                    temp[:] = diff[:]
                    temp[:, j] = 1.0
                    temp[:, k] = 1.0
                    Di[:, j] += wb[j] * np.prod(temp, axis=1)
    else:
        Di = None

    return Li, Di



import openmdao.api as om
from .aero_forces_comp import AeroForcesComp
from .interpND_aero_coef_comp import InterpNDAeroCoeffComp, setup_surrogates_all


class AerodynamicsGroup(om.Group):
    """
    The purpose of the Aerodynamics is to compute the lift and
    drag forces on the aircraft.
    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        n = self.options['num_nodes']

        interpND_CL, interpND_CD, interpND_CM, interp_num = setup_surrogates_all()

        self.add_subsystem(name='aero_coef_comp',
                           subsys=InterpNDAeroCoeffComp(vec_size=n, interpND_CL=interpND_CL, interpND_CD=interpND_CD,
                                                        interpND_CM=interpND_CM,
                                                        interp_num=interp_num),
                           promotes_inputs=[('M', 'mach'), 'alpha', ('h', 'alt'), 'eta'],
                           promotes_outputs=['CL', 'CD', 'CM'])

        self.add_subsystem(name='aero_forces_comp',
                           subsys=AeroForcesComp(num_nodes=n),
                           promotes_inputs=['q', 'S', 'CL', 'CD'],
                           promotes_outputs=['L', 'D'])



import numpy as np
import openmdao.api as om


class RangeRateComp(om.ExplicitComponent):
    """
    Calculates range rate based on true airspeed and flight path angle.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('TAS', val=np.ones(nn), desc='True airspeed', units='m/s')

        self.add_input('gam', val=np.zeros(nn), desc='Flight path angle', units='rad')

        self.add_output('dXdt:range', val=np.ones(nn), desc='Velocity along the ground (no wind)',
                        units='m/s')

        # Setup partials
        ar = np.arange(self.options['num_nodes'])
        self.declare_partials(of='*', wrt='*', dependent=False)
        self.declare_partials(of='dXdt:range', wrt='TAS', rows=ar, cols=ar)
        self.declare_partials(of='dXdt:range', wrt='gam', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        TAS = inputs['TAS']
        gam = inputs['gam']
        outputs['dXdt:range'] = TAS*np.cos(gam)

    def compute_partials(self, inputs, partials):
        TAS = inputs['TAS']
        gam = inputs['gam']

        partials['dXdt:range', 'TAS'] = np.cos(gam)
        partials['dXdt:range', 'gam'] = -TAS * np.sin(gam)



import numpy as np
import openmdao.api as om

from ...utils.lgl import lgl
from ...utils.misc import get_rate_units


class VandermondeControlInterpComp(om.ExplicitComponent):
    """
    A component which interpolates control values in 1D using Vandermonde interpolation.

    Takes training values for control variables at given _input_ nodes,
    broadcasts them to _discretization_ nodes, and then interpolates the discretization values
    to provide a control variable at a given segment tau or phase tau.

    For dynamic controls, the current segment is given as a discrete input and the interpolation is
    a smooth polynomial along the given segment.

    OpenMDAO assumes sizes of variables at setup time, and we don't want to need to change the
    size of the control input nodes when we evaluate different segments. Instead, this component
    will take in the control values of all segments and internally use the appropriate one.

    Parameters
    ----------
    grid_data : GridData
        A GridData instance that details information on how the control input and discretization
        nodes are layed out.
    control_options : dict of {str: ControlOptionsDictionary}
        A mapping that maps the name of each control to a ControlOptionsDictionary of its options.
    time_units : str
        The time units pertaining to the control rates.
    standalone_mode : bool
        If True, this component runs its configuration steps during setup. This is useful for
        unittests in which the component does not exist in a larger group.
    **kwargs
        Keyword arguments passed to ExplicitComponent.
    """
    def __init__(self, grid_data, control_options=None,
                 time_units=None, standalone_mode=False, **kwargs):
        self._grid_data = grid_data
        self._control_options = {} if control_options is None else control_options
        self._time_units = time_units
        self._standalone_mode = standalone_mode

        # Storage for the Vandermonde matrix and its inverse for each control
        self._V_hat = {}
        self._V_hat_inv = {}

        # Storage for factors used in the derivatives of Vandermonde matrices.
        self._fac = {}

        # Cache formatted strings: { control_name : (input_name, output_name) }
        self._control_io_names = {}

        # The Lagrange interpolation matrix L_id maps control values as given at the input nodes
        # to values at the discretization nodes.
        num_disc_nodes = grid_data.subset_num_nodes['control_disc']
        num_input_nodes = grid_data.subset_num_nodes['control_input']
        self._L_id = np.zeros((num_disc_nodes, num_input_nodes), dtype=float)
        self._L_id[np.arange(num_disc_nodes, dtype=int),
                   self._grid_data.input_maps['dynamic_control_input_to_disc']] = 1.0

        super().__init__(**kwargs)

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('segment_index', types=int, desc='index of the current segment')
        self.options.declare('vec_size', types=int, default=1,
                             desc='number of points at which the control will be evaluated. This is not'
                                  'necessarily the same as the number of nodes in the GridData.')
        self.options.declare('compute_derivs', types=bool, default=True,
                             desc='Set to True if the interpolant needs to also compute the derivatives '
                             'of the outputs, otherwise False. This should be set to False for simulation mode '
                             'and true when using ExplicitShooting for optimization')

    def _configure_controls(self):
        vec_size = self.options['vec_size']
        gd = self._grid_data

        self._V_hat = {}
        self._V_hat_inv = {}
        self._disc_node_idxs_by_segment = []
        self._input_node_idxs_by_segment = []

        if not self._control_options:
            return

        first_disc_node_in_seg = 0

        for seg_idx in range(gd.num_segments):
            # Number of control discretization nodes per segment
            ncdnps = gd.subset_num_nodes_per_segment['control_disc'][seg_idx]
            ar_control_disc_nodes = np.arange(ncdnps, dtype=int)
            disc_idxs_in_seg = first_disc_node_in_seg + ar_control_disc_nodes
            first_disc_node_in_seg += ncdnps

            # The indices of the discretization node u vector pertaining to the given segment
            self._disc_node_idxs_by_segment.append(disc_idxs_in_seg)

            # The indices of the input u vector pertaining to the given segment
            self._input_node_idxs_by_segment.append(gd.input_maps['dynamic_control_input_to_disc'][disc_idxs_in_seg])

            # Indices of the control disc nodes belonging to the current segment
            control_disc_seg_idxs = gd.subset_segment_indices['control_disc'][seg_idx]

            # Segment tau values for the control disc nodes in the phase
            control_disc_stau = gd.node_stau[gd.subset_node_indices['control_disc']]

            # Segment tau values for the control disc nodes in the given segment
            control_disc_seg_stau = control_disc_stau[control_disc_seg_idxs[0]:
                                                      control_disc_seg_idxs[1]]

            seg_control_order = gd.transcription_order[seg_idx] - 1
            if seg_control_order not in self._V_hat:
                self._V_hat[seg_control_order] = np.vander(control_disc_seg_stau, increasing=True)
                self._V_hat_inv[seg_control_order] = np.linalg.inv(self._V_hat[seg_control_order])
            if seg_control_order + 1 not in self._fac:
                self._fac[seg_control_order + 1] = np.arange(seg_control_order + 1, dtype=int)

        num_uhat_nodes = gd.subset_num_nodes['control_input']
        ar = np.arange(vec_size, dtype=int)
        for control_name, options in self._control_options.items():
            if options['control_type'] == 'full':
                shape = options['shape']
                units = options['units']
                input_name = f'controls:{control_name}'
                output_name = f'control_values:{control_name}'
                rate_name = f'control_rates:{control_name}_rate'
                rate2_name = f'control_rates:{control_name}_rate2'
                rate_units = get_rate_units(units, self._time_units)
                rate2_units = get_rate_units(units, self._time_units, deriv=2)
                uhat_shape = (num_uhat_nodes,) + shape
                output_shape = (vec_size,) + shape
                self.add_input(input_name, shape=uhat_shape, units=units)
                self.add_output(output_name, shape=output_shape, units=units)
                self.add_output(rate_name, shape=output_shape, units=rate_units)
                self.add_output(rate2_name, shape=output_shape, units=rate2_units)
                self._control_io_names[control_name] = (input_name, output_name, rate_name, rate2_name)
                self.declare_partials(of=output_name, wrt=input_name)
                self.declare_partials(of=output_name, wrt='stau', rows=ar, cols=ar)
                self.declare_partials(of=rate_name, wrt=input_name)
                self.declare_partials(of=rate_name, wrt='stau', rows=ar, cols=ar)
                self.declare_partials(of=rate_name, wrt='dstau_dt')
                self.declare_partials(of=rate2_name, wrt=input_name)
                self.declare_partials(of=rate2_name, wrt='stau', rows=ar, cols=ar)
                self.declare_partials(of=rate2_name, wrt='dstau_dt')
            else:
                order = options['order']
                shape = options['shape']
                units = options['units']
                input_name = f'controls:{control_name}'
                output_name = f'control_values:{control_name}'
                rate_name = f'control_rates:{control_name}_rate'
                rate2_name = f'control_rates:{control_name}_rate2'
                rate_units = get_rate_units(units, self._time_units)
                rate2_units = get_rate_units(units, self._time_units, deriv=2)
                input_shape = (order + 1,) + shape
                output_shape = (vec_size,) + shape
                self.add_input(input_name, shape=input_shape, units=units)
                self.add_output(output_name, shape=output_shape, units=units)
                self.add_output(rate_name, shape=output_shape, units=rate_units)
                self.add_output(rate2_name, shape=output_shape, units=rate2_units)
                self._control_io_names[control_name] = (input_name, output_name, rate_name, rate2_name)
                self.declare_partials(of=output_name, wrt=input_name)
                self.declare_partials(of=output_name, wrt='ptau', rows=ar, cols=ar)
                self.declare_partials(of=rate_name, wrt=input_name)
                self.declare_partials(of=rate_name, wrt='ptau', rows=ar, cols=ar)
                self.declare_partials(of=rate_name, wrt='t_duration')
                self.declare_partials(of=rate2_name, wrt=input_name)
                self.declare_partials(of=rate2_name, wrt='ptau', rows=ar, cols=ar)
                self.declare_partials(of=rate2_name, wrt='t_duration')

                if order not in self._V_hat:
                    pc_disc_seg_ptau, _ = lgl(order + 1)
                    self._V_hat[order] = np.vander(pc_disc_seg_ptau, increasing=True)
                    self._V_hat_inv[order] = np.linalg.inv(self._V_hat[order])
                if order + 1 not in self._fac:
                    self._fac[order + 1] = np.arange(order + 1, dtype=int)

    def setup(self):
        """
        Perform the I/O creation if operating in _standalone_mode.
        """
        if self._standalone_mode:
            self.configure_io()

    def set_segment_index(self, idx, **kwargs):
        """
        Set the active segment index for control interpolation.

        Parameters
        ----------
        idx : int
            The index of the segment in which the controls are to be interpolated.
        **kwargs : dict, optional
            Keyword arguments that make this interpolant call-compatible with the BarycentricLagrangeInterpolant.
        """
        self.options['segment_index'] = idx

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine shape and units for the controls.
        """
        vec_size = self.options['vec_size']

        self._V_hat = {}
        self._V_hat_inv = {}

        # self.add_discrete_input('segment_index', val=0, desc='index of the segment')
        self.add_input('stau', shape=(vec_size,), units=None)
        self.add_input('dstau_dt', val=1.0, units=f'1/{self._time_units}')
        self.add_input('t_duration', val=1.0, units=self._time_units)
        self.add_input('ptau', shape=(vec_size,), units=None)

        self._configure_controls()

    def _dvander(self, v):
        """
        Return the derivative of a Vandermonde matrix wrt the independent variable _in increasing order_.

        Parameters
        ----------
        v : np.array
            The Vandermonde matrix for which the derivatives are requested.

        Returns
        -------
        dV : np.array
            The derivative of v with respect to the independent variable.
        dv2 : np.array
            The second derivative of v wrt the independent variable.
        dv3 : np.array
            The third derivative of v wrt the independent variable.
        """
        p, n = v.shape
        dv = np.zeros_like(v)
        dv2 = dv.copy()
        if self.options['compute_derivs']:
            dv3 = dv.copy()
        else:
            dv3 = None
        dv[:, 1:] = v[:, :-1]
        dv2[:, 2:] = v[:, :-2]

        if self.options['compute_derivs']:
            dv3[:, 3:] = v[:, :-3]

        fac = self._fac[n]
        fac2 = fac[:-1]
        fac3 = fac[:-2]

        dv[:, :] = dv * fac[np.newaxis, :]
        dv2[:, 1:] = dv2[:, 1:] * fac2[np.newaxis, :] * fac[np.newaxis, 1:]
        if self.options['compute_derivs']:
            dv3[:, 2:] = dv3[:, 2:] * fac3[np.newaxis, :] * fac2[np.newaxis, 1:] * fac[np.newaxis, 2:]
        return dv, dv2, dv3

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Compute interpolated control values and rates.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        discrete_inputs : `Vector`
            `Vector` containing discrete_inputs.
        discrete_outputs : `Vector`
            `Vector` containing discrete_outputs.
        """
        seg_idx = self.options['segment_index']
        n = self._grid_data.transcription_order[seg_idx]
        stau = inputs['stau']
        dstau_dt = inputs['dstau_dt']
        ptau = inputs['ptau']
        dptau_dt = 2 / inputs['t_duration']

        if self._control_options:
            seg_order = n - 1
            disc_node_idxs = self._disc_node_idxs_by_segment[seg_idx]
            input_node_idxs = self._input_node_idxs_by_segment[seg_idx]
            V_stau = np.vander(stau, N=n, increasing=True)
            dV_stau, dV2_stau, _ = self._dvander(V_stau)

            L_seg = self._L_id[disc_node_idxs[0]:disc_node_idxs[0] + len(disc_node_idxs),
                               input_node_idxs[0]:input_node_idxs[0] + len(input_node_idxs)]

            for control_name, options in self._control_options.items():
                if options['control_type'] == 'full':
                    input_name, output_name, rate_name, rate2_name = self._control_io_names[control_name]
                    u_hat = np.dot(L_seg, inputs[input_name][input_node_idxs])
                    a = np.atleast_2d(self._V_hat_inv[seg_order] @ u_hat)
                    outputs[output_name] = V_stau @ a
                    outputs[rate_name] = dstau_dt * (dV_stau @ a)
                    outputs[rate2_name] = dstau_dt**2 * (dV2_stau @ a)
                else:
                    input_name, output_name, rate_name, rate2_name = self._control_io_names[control_name]
                    order = options['order']
                    V_ptau = np.vander(ptau, N=order+1, increasing=True)
                    dV_ptau, dV2_ptau, _ = self._dvander(V_ptau)
                    a = np.atleast_2d(self._V_hat_inv[order] @ inputs[input_name])
                    outputs[output_name] = V_ptau @ a
                    outputs[rate_name] = dptau_dt * (dV_ptau @ a)
                    outputs[rate2_name] = dptau_dt**2 * (dV2_ptau @ a)

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        """
        Compute derivatives of interpolated control values and rates wrt the inputs.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        discrete_inputs : Vector
            Unscaled, discrete input variables keyed by variable name.
        """
        seg_idx = self.options['segment_index']
        n = self._grid_data.transcription_order[seg_idx]
        stau = inputs['stau'].real
        dstau_dt = inputs['dstau_dt'].real
        ptau = inputs['ptau'].real
        t_duration = inputs['t_duration'].real
        dptau_dt = 2.0 / t_duration
        ddptau_dt_dtduration = -2.0 / t_duration**2

        if self._control_options:
            u_idxs = self._input_node_idxs_by_segment[seg_idx]
            seg_order = self._grid_data.transcription_order[seg_idx] - 1

            V_stau = np.vander(stau, N=n, increasing=True)
            dV_stau, dV2_stau, dV3_stau = self._dvander(V_stau)

            disc_node_idxs = self._disc_node_idxs_by_segment[seg_idx]
            input_node_idxs = self._input_node_idxs_by_segment[seg_idx]

            L_seg = self._L_id[disc_node_idxs[0]:disc_node_idxs[0] + len(disc_node_idxs),
                               input_node_idxs[0]:input_node_idxs[0] + len(input_node_idxs)]

            for control_name, options in self._control_options.items():
                if options['control_type'] == 'full':
                    input_name, output_name, rate_name, rate2_name = self._control_io_names[control_name]

                    u_hat = np.dot(L_seg, inputs[input_name][input_node_idxs].real)
                    a = self._V_hat_inv[seg_order] @ u_hat

                    da_duhat = self._V_hat_inv[seg_order] @ L_seg
                    dV_a = dV_stau @ a
                    dV2_a = dV2_stau @ a
                    dV3_a = dV3_stau @ a

                    partials[output_name, input_name][...] = 0.0
                    partials[output_name, input_name][..., u_idxs] = V_stau @ da_duhat
                    partials[output_name, 'stau'] = dV_a.ravel()

                    pudot_pa = dstau_dt * dV_stau
                    pa_puhat = self._V_hat_inv[seg_order]
                    partials[rate_name, input_name][...] = 0.0
                    partials[rate_name, input_name][..., u_idxs] = pudot_pa @ pa_puhat
                    partials[rate_name, 'dstau_dt'][...] = dV_a
                    partials[rate_name, 'stau'][...] = dV2_a.ravel()

                    pu2dot_pa = dstau_dt**2 * dV2_stau
                    partials[rate2_name, input_name][...] = 0.0
                    partials[rate2_name, input_name][..., u_idxs] = pu2dot_pa @ pa_puhat
                    partials[rate2_name, 'dstau_dt'][...] = 2 * dstau_dt * dV2_a
                    partials[rate2_name, 'stau'][...] = dV3_a.ravel()

                else:
                    input_name, output_name, rate_name, rate2_name = self._control_io_names[control_name]
                    order = options['order']

                    V_ptau = np.vander(ptau, N=order+1, increasing=True)
                    dV_ptau, dV2_ptau, dV3_ptau = self._dvander(V_ptau)

                    u_hat = inputs[input_name].real
                    a = self._V_hat_inv[order] @ u_hat

                    dV_a = dV_ptau @ a
                    dV2_a = dV2_ptau @ a
                    dV3_a = dV3_ptau @ a

                    da_duhat = self._V_hat_inv[order]

                    partials[output_name, input_name][...] = V_ptau @ da_duhat
                    partials[output_name, 'ptau'][...] = dV_a.ravel()

                    pudot_pa = dptau_dt * dV_ptau
                    pa_puhat = self._V_hat_inv[order]
                    partials[rate_name, input_name][...] = pudot_pa @ pa_puhat
                    partials[rate_name, 't_duration'][...] = ddptau_dt_dtduration * dV_a
                    partials[rate_name, 'ptau'][...] = dptau_dt * dV2_a.ravel()

                    pu2dot_pa = dptau_dt ** 2 * dV2_ptau
                    partials[rate2_name, input_name][...] = pu2dot_pa @ pa_puhat
                    partials[rate2_name, 't_duration'][...] = 2 * dptau_dt * ddptau_dt_dtduration * dV2_a
                    partials[rate2_name, 'ptau'][...] = dptau_dt**2 * dV3_a.ravel()



from .analytic.analytic import Analytic
from .explicit_shooting import ExplicitShooting
from .pseudospectral.gauss_lobatto import GaussLobatto
from .pseudospectral.radau_pseudospectral import Radau
from .pseudospectral.birkhoff import Birkhoff
from .picard_shooting.picard_shooting import PicardShooting



import unittest

from numpy.testing import assert_almost_equal

from dymos.utils.birkhoff import birkhoff_matrix
from dymos.utils.cgl import cgl, clenshaw_curtis
from dymos.utils.lgl import lgl


class TestBirkhoffMatrix(unittest.TestCase):

    def test_birkhoff_matrix_lgl(self):
        x, w = lgl(9)
        B = birkhoff_matrix(x, w, grid_type='lgl')
        assert_almost_equal(B[-1, :], w)

    def test_birkhoff_matrix_cgl(self):
        x, w = cgl(9)
        B = birkhoff_matrix(x, w, grid_type='cgl')
        _, wB = clenshaw_curtis(9)
        assert_almost_equal(B[-1, :], wB)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import openmdao.api as om
import dymos as dm


def make_balanced_field_length_problem(ode_class, tx):
    """
    Create a balanced field length problem and optionally set default values into it.

    Parameters
    ----------
    ode_class : System class
        The Dymos ODE System class.
    tx_class : Transcription
        Transcription to use.

    Returns
    -------
    _type_
        _description_
    """
    p = om.Problem()

    p.driver = om.pyOptSparseDriver()
    p.driver.declare_coloring()

    # Use IPOPT if available, with fallback to SLSQP
    p.driver.options['optimizer'] = 'IPOPT'
    p.driver.options['print_results'] = True

    p.driver.opt_settings['print_level'] = 0
    p.driver.opt_settings['mu_strategy'] = 'adaptive'

    p.driver.opt_settings['bound_mult_init_method'] = 'mu-based'
    p.driver.opt_settings['mu_init'] = 0.01
    p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'

    # First Phase: Brake release to V1 - both engines operable
    br_to_v1 = dm.Phase(ode_class=ode_class, transcription=tx,
                        ode_init_kwargs={'mode': 'runway'})
    br_to_v1.set_time_options(fix_initial=True, duration_bounds=(1, 1000), duration_ref=10.0)
    br_to_v1.add_state('r', fix_initial=True, lower=0, ref=1000.0, defect_ref=1000.0)
    br_to_v1.add_state('v', fix_initial=True, lower=0, ref=100.0, defect_ref=100.0)
    br_to_v1.add_parameter('alpha', val=0.0, opt=False, units='deg')
    br_to_v1.add_timeseries_output('*')

    # Second Phase: Rejected takeoff at V1 - no engines operable
    rto = dm.Phase(ode_class=ode_class, transcription=tx,
                   ode_init_kwargs={'mode': 'runway'})
    rto.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
    rto.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
    rto.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
    rto.add_parameter('alpha', val=0.0, opt=False, units='deg')
    rto.add_timeseries_output('*')

    # Third Phase: V1 to Vr - single engine operable
    v1_to_vr = dm.Phase(ode_class=ode_class, transcription=tx,
                        ode_init_kwargs={'mode': 'runway'})
    v1_to_vr.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
    v1_to_vr.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
    v1_to_vr.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
    v1_to_vr.add_parameter('alpha', val=0.0, opt=False, units='deg')
    v1_to_vr.add_timeseries_output('*')

    # Fourth Phase: Rotate - single engine operable
    rotate = dm.Phase(ode_class=ode_class, transcription=tx,
                      ode_init_kwargs={'mode': 'runway'})
    rotate.set_time_options(fix_initial=False, duration_bounds=(1.0, 5), duration_ref=1.0)
    rotate.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
    rotate.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
    rotate.add_control('alpha', order=1, opt=True, units='deg', lower=0, upper=10, ref=10,
                       val=[0, 10], control_type='polynomial')
    rotate.add_timeseries_output('*')

    # Fifth Phase: Climb to target speed and altitude at end of runway.
    climb = dm.Phase(ode_class=ode_class, transcription=tx,
                     ode_init_kwargs={'mode': 'climb'})
    climb.set_time_options(fix_initial=False, duration_bounds=(1, 100), duration_ref=1.0)
    climb.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
    climb.add_state('h', fix_initial=True, lower=0, ref=1.0, defect_ref=1.0)
    climb.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
    climb.add_state('gam', fix_initial=True, lower=0, ref=0.05, defect_ref=0.05)
    climb.add_control('alpha', opt=True, units='deg', lower=-10, upper=15, ref=10)
    climb.add_timeseries_output('*')

    # Instantiate the trajectory and add phases
    traj = dm.Trajectory()
    p.model.add_subsystem('traj', traj)
    traj.add_phase('br_to_v1', br_to_v1)
    traj.add_phase('rto', rto)
    traj.add_phase('v1_to_vr', v1_to_vr)
    traj.add_phase('rotate', rotate)
    traj.add_phase('climb', climb)

    all_phases = ['br_to_v1', 'v1_to_vr', 'rto', 'rotate', 'climb']
    groundroll_phases = ['br_to_v1', 'v1_to_vr', 'rto', 'rotate']

    # Add parameters common to multiple phases to the trajectory
    traj.add_parameter('m', val=174200., opt=False, units='lbm',
                       desc='aircraft mass',
                       targets={phase: ['m'] for phase in all_phases})

    # Handle parameters which change from phase to phase.
    traj.add_parameter('T_nominal', val=27000 * 2, opt=False, units='lbf', static_target=True,
                       desc='nominal aircraft thrust',
                       targets={'br_to_v1': ['T']})

    traj.add_parameter('T_engine_out', val=27000, opt=False, units='lbf', static_target=True,
                       desc='thrust under a single engine',
                       targets={'v1_to_vr': ['T'], 'rotate': ['T'], 'climb': ['T']})

    traj.add_parameter('T_shutdown', val=0.0, opt=False, units='lbf', static_target=True,
                       desc='thrust when engines are shut down for rejected takeoff',
                       targets={'rto': ['T']})

    traj.add_parameter('mu_r_nominal', val=0.03, opt=False, units=None, static_target=True,
                       desc='nominal runway friction coefficient',
                       targets={'br_to_v1': ['mu_r'], 'v1_to_vr': ['mu_r'],  'rotate': ['mu_r']})

    traj.add_parameter('mu_r_braking', val=0.3, opt=False, units=None, static_target=True,
                       desc='runway friction coefficient under braking',
                       targets={'rto': ['mu_r']})

    traj.add_parameter('h_runway', val=0., opt=False, units='ft',
                       desc='runway altitude',
                       targets={phase: ['h'] for phase in groundroll_phases})

    # Here we're omitting some constants that are common throughout all phases for the sake of brevity.
    # Their correct defaults are specified in add_input calls to `wrap_ode_func`.

    # Standard "end of first phase to beginning of second phase" linkages
    # Alpha changes from being a parameter in v1_to_vr to a polynomial control
    # in rotate, to a dynamic control in `climb`.
    traj.link_phases(['br_to_v1', 'v1_to_vr'], vars=['time', 'r', 'v'])
    traj.link_phases(['v1_to_vr', 'rotate'], vars=['time', 'r', 'v', 'alpha'])
    traj.link_phases(['rotate', 'climb'], vars=['time', 'r', 'v', 'alpha'])
    traj.link_phases(['br_to_v1', 'rto'], vars=['time', 'r', 'v'])

    # Less common "final value of r must match at ends of two phases".
    traj.add_linkage_constraint(phase_a='rto', var_a='r', loc_a='final',
                                phase_b='climb', var_b='r', loc_b='final',
                                ref=1000)

    # Define the constraints and objective for the optimal control problem
    v1_to_vr.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.2, ref=100)

    rto.add_boundary_constraint('v', loc='final', equals=0., ref=100, linear=True)

    rotate.add_boundary_constraint('F_r', loc='final', equals=0, ref=100000)

    climb.add_boundary_constraint('h', loc='final', equals=35, ref=35, units='ft', linear=True)
    climb.add_boundary_constraint('gam', loc='final', equals=5, ref=5, units='deg', linear=True)
    climb.add_path_constraint('gam', lower=0, upper=5, ref=5, units='deg')
    climb.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.25, ref=1.25)

    rto.add_objective('r', loc='final', ref=1000.0)

    #
    # Setup the problem and set the initial guess
    #
    p.setup(check=True)

    br_to_v1.set_time_val(initial=0.0, duration=35.0)
    br_to_v1.set_state_val('r', [0, 2500.0])
    br_to_v1.set_state_val('v', [0.0001, 100.0])
    br_to_v1.set_parameter_val('alpha', 0.0, units='deg')

    v1_to_vr.set_time_val(initial=35.0, duration=35.0)
    v1_to_vr.set_state_val('r', [2500, 300.0])
    v1_to_vr.set_state_val('v', [100, 110.0])
    v1_to_vr.set_parameter_val('alpha', 0.0, units='deg')

    rto.set_time_val(initial=35.0, duration=1.0)
    rto.set_state_val('r', [2500, 5000.0])
    rto.set_state_val('v', [110, 0.0001])
    rto.set_parameter_val('alpha', 0.0, units='deg')

    rotate.set_time_val(initial=35.0, duration=5.0)
    rotate.set_state_val('r', [1750, 1800.0])
    rotate.set_state_val('v', [80, 85.0])
    rotate.set_control_val('alpha', 0.0, units='deg')

    climb.set_time_val(initial=30.0, duration=20.0)
    climb.set_state_val('r', [5000, 5500.0], units='ft')
    climb.set_state_val('v', [160, 170.0], units='kn')
    climb.set_state_val('h', [0.0, 35.0], units='ft')
    climb.set_state_val('gam', [0.0, 5.0], units='deg')
    climb.set_control_val('alpha', 5.0, units='deg')

    return p



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

from dymos.utils.cgl import cgl

# Known solutions
x_i = {2: [-1.0, 1.0],
       3: [-1.0, 0.0, 1.0],
       4: [-1.0, -0.5, 0.5, 1.0],
       5: [-1.0, -1/np.sqrt(2), 0.0, 1/np.sqrt(2), 1.0]}

w_i = {2: [np.pi/2, np.pi/2],
       3: [np.pi/4, np.pi/2, np.pi/4],
       4: [np.pi/6, np.pi/3, np.pi/3, np.pi/6],
       5: [np.pi/8, np.pi/4, np.pi/4, np.pi/4, np.pi/8]}


class TestCGL(unittest.TestCase):

    def test_nodes_and_weights_2(self):
        x_2, w_2 = cgl(2)
        assert_almost_equal(x_2, x_i[2], decimal=6)
        assert_almost_equal(w_2, w_i[2], decimal=6)

    def test_nodes_and_weights_3(self):
        x_3, w_3 = cgl(3)
        assert_almost_equal(x_3, x_i[3], decimal=6)
        assert_almost_equal(w_3, w_i[3], decimal=6)

    def test_nodes_and_weights_4(self):
        x_4, w_4 = cgl(4)
        assert_almost_equal(x_4, x_i[4], decimal=6)
        assert_almost_equal(w_4, w_i[4], decimal=6)

    def test_nodes_and_weights_5(self):
        x_5, w_5 = cgl(5)
        assert_almost_equal(x_5, x_i[5], decimal=6)
        assert_almost_equal(w_5, w_i[5], decimal=6)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import numpy as np

try:
    import matplotlib
except ImportError:
    matplotlib = None

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs
from dymos.utils.misc import om_version
from dymos.utils.testing_utils import assert_timeseries_near_equal
from dymos.utils.misc import _unspecified


class BrachistochroneRateTargetODE(om.ExplicitComponent):
    #
    # The following dictionaries provide a way of 'tagging' the Brachistochrone ODE with
    # information about states and parameters that can be accessed from Phase.
    #
    # In this case these are class attributes, but the choice of whether or not to tie
    # this information to the ODE itself (and how to do so) is entirely up to the user.
    #
    # In a dynamic ODE model these might be instance attributes whose values vary depending on
    # the arguments to the instantiation.
    #
    states = {'x': {'rate_source': 'xdot',
                    'units': 'm'},
              'y': {'rate_source': 'ydot',
                    'units': 'm'},
              'v': {'rate_source': 'vdot',
                    'units': 'm/s'}}

    parameters = {'theta': {'units': 'rad'},
                  'g': {'units': 'm/s**2'}}

    def initialize(self):
        self.options.declare('num_nodes', types=int)
        self.options.declare('static_gravity', types=(bool,), default=False,
                             desc='If True, treat gravity as a static (scalar) input, rather than '
                                  'having different values at each node.')
        self.options.declare('control_name', values=('theta', 'int_theta_rate'), default='int_theta_rate')

    def setup(self):
        nn = self.options['num_nodes']
        control_name = self.options['control_name']
        g_default_val = 9.80665 if self.options['static_gravity'] else 9.80665 * np.ones(nn)

        # Inputs
        self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')

        self.add_input('g', val=g_default_val, desc='grav. acceleration', units='m/s/s')

        # Note, kind of strange for this to be named theta_rate, but this is just a demonstration that
        # we can connect to a control rate source in dymos.
        self.add_input(control_name, val=np.ones(nn), desc='angle of wire', units='rad')

        self.add_output('xdot', val=np.zeros(nn), desc='velocity component in x', units='m/s')

        self.add_output('ydot', val=np.zeros(nn), desc='velocity component in y', units='m/s')

        self.add_output('vdot', val=np.zeros(nn), desc='acceleration magnitude', units='m/s**2')

        self.add_output('check', val=np.zeros(nn),
                        desc='check solution: v/sin(theta) = constant',
                        units='m/s')

        # Setup partials
        arange = np.arange(self.options['num_nodes'])
        self.declare_partials(of='vdot', wrt=control_name, rows=arange, cols=arange)

        self.declare_partials(of='xdot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='xdot', wrt=control_name, rows=arange, cols=arange)

        self.declare_partials(of='ydot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='ydot', wrt=control_name, rows=arange, cols=arange)

        self.declare_partials(of='check', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='check', wrt=control_name, rows=arange, cols=arange)

        if self.options['static_gravity']:
            c = np.zeros(self.options['num_nodes'])
            self.declare_partials(of='vdot', wrt='g', rows=arange, cols=c)
        else:
            self.declare_partials(of='vdot', wrt='g', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        u_name = self.options['control_name']
        theta = inputs[u_name]
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        outputs['vdot'] = g * cos_theta
        outputs['xdot'] = v * sin_theta
        outputs['ydot'] = -v * cos_theta
        outputs['check'] = v / sin_theta

    def compute_partials(self, inputs, partials):
        u_name = self.options['control_name']
        theta = inputs[u_name]
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        partials['vdot', 'g'] = cos_theta
        partials['vdot', u_name] = -g * sin_theta

        partials['xdot', 'v'] = sin_theta
        partials['xdot', u_name] = v * cos_theta

        partials['ydot', 'v'] = -cos_theta
        partials['ydot', u_name] = v * sin_theta

        partials['check', 'v'] = 1 / sin_theta
        partials['check', u_name] = -v * cos_theta / sin_theta ** 2


@use_tempdirs
class TestBrachistochroneControlRateTargets(unittest.TestCase):

    def test_brachistochrone_control_rate_targets(self):

        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm

        transcriptions = {'gauss-lobatto': dm.GaussLobatto(num_segments=10),
                          'radau': dm.Radau(num_segments=10),
                          'birkhoff': dm.Birkhoff(num_nodes=30)}

        for tx_name, tx in transcriptions.items():
            for control_target_method in ('implicit', 'explicit'):
                for control_type in ('full', 'polynomial'):

                    with self.subTest(f'{tx_name=} {control_target_method=} {control_type=}'):

                        p = om.Problem(model=om.Group())
                        p.driver = om.ScipyOptimizeDriver()
                        p.driver.declare_coloring()

                        traj = dm.Trajectory()

                        ode_kwargs = {'control_name': 'int_theta_rate' if control_target_method == 'implicit' else 'theta'}

                        phase = dm.Phase(ode_class=BrachistochroneRateTargetODE,
                                         ode_init_kwargs=ode_kwargs,
                                         transcription=tx)

                        traj.add_phase('phase0', phase)

                        p.model.add_subsystem('traj0', traj)

                        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

                        phase.add_state('x', rate_source='xdot',
                                        units='m',
                                        fix_initial=True, fix_final=True, solve_segments=False)

                        phase.add_state('y', rate_source='ydot',
                                        units='m',
                                        fix_initial=True, fix_final=True, solve_segments=False)

                        phase.add_state('v', rate_source='vdot',
                                        units='m/s',
                                        fix_initial=True, fix_final=False, solve_segments=False)

                        phase.add_control('int_theta', lower=0.0, upper=None, fix_initial=True,
                                          rate_targets=_unspecified if control_target_method == 'implicit' else ['theta'],
                                          control_type=control_type, order=7)

                        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

                        # Minimize time at the end of the phase
                        phase.add_objective('time', loc='final', scaler=10)

                        p.model.linear_solver = om.DirectSolver()

                        phase.timeseries_options['include_control_rates'] = True

                        p.setup()

                        phase.set_time_val(initial=0.0, duration=2.0)

                        phase.set_state_val('x', [0, 10])
                        phase.set_state_val('y', [10, 5])
                        phase.set_state_val('v', [0, 9.9])
                        phase.set_control_val('int_theta', [0, 100], units='deg*s')

                        p.run_model()

                        # Solve for the optimal trajectory
                        dm.run_problem(p, simulate=True, make_plots=True)

                        sol_db = 'dymos_solution.db'
                        sim_db = 'dymos_simulation.db'
                        if om_version()[0] > (3, 34, 2):
                            sol_db = p.get_outputs_dir() / sol_db
                            sim_db = traj.sim_prob.get_outputs_dir() / sim_db

                        sol_case = om.CaseReader(sol_db).get_case('final')
                        sim_case = om.CaseReader(sim_db).get_case('final')

                        t_sol = sol_case.get_val('traj0.phase0.timeseries.time')
                        x_sol = sol_case.get_val('traj0.phase0.timeseries.x')
                        y_sol = sol_case.get_val('traj0.phase0.timeseries.y')
                        v_sol = sol_case.get_val('traj0.phase0.timeseries.v')
                        theta_sol = sol_case.get_val('traj0.phase0.timeseries.int_theta_rate')

                        t_sim = sim_case.get_val('traj0.phase0.timeseries.time')
                        x_sim = sim_case.get_val('traj0.phase0.timeseries.x')
                        y_sim = sim_case.get_val('traj0.phase0.timeseries.y')
                        v_sim = sim_case.get_val('traj0.phase0.timeseries.v')
                        theta_sim = sim_case.get_val('traj0.phase0.timeseries.int_theta_rate')

                        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, rel_tolerance=5.0E-3, abs_tolerance=0.01)
                        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, rel_tolerance=5.0E-3, abs_tolerance=0.01)
                        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, rel_tolerance=5.0E-3, abs_tolerance=0.01)
                        assert_timeseries_near_equal(t_sol, theta_sol, t_sim, theta_sim, rel_tolerance=5.0E-3, abs_tolerance=0.01)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np


def linewidth_from_data_units(linewidth, axis, reference='y'):
    """
    Convert a linewidth in data units to linewidth in points.

    Parameters
    ----------
    linewidth: float
        Linewidth in data units of the respective reference-axis
    axis: matplotlib axis
        The axis which is used to extract the relevant transformation
        data (data limits and size must not change afterwards)
    reference: string
        The axis that is taken as a reference for the data width.
        Possible values: 'x' and 'y'. Defaults to 'y'.

    Returns
    -------
    linewidth: float
        Linewidth in points
    """
    fig = axis.get_figure()
    if reference == 'x':
        length = fig.bbox_inches.width * axis.get_position().width
        value_range = np.diff(axis.get_xlim())
    elif reference == 'y':
        length = fig.bbox_inches.height * axis.get_position().height
        value_range = np.diff(axis.get_ylim())
    # Convert length to points
    length *= 72
    # Scale linewidth to value range
    return linewidth * (length / value_range)



import unittest

from openmdao.utils.testing_utils import use_tempdirs, set_env_vars
import dymos as dm
from dymos.examples.brachistochrone.test.ex_brachistochrone import brachistochrone_min_time


@use_tempdirs
class TestOptions(unittest.TestCase):

    @set_env_vars(CI='0')
    def test_include_check_partials_false_radau(self):
        dm.options['include_check_partials'] = False
        p = brachistochrone_min_time(transcription='radau-ps', compressed=False,
                                     run_driver=False, force_alloc_complex=True)
        cpd = p.check_partials(out_stream=None)
        self.assertSetEqual(set(cpd.keys()), {'traj0.phases.phase0.rhs_all'})

    @set_env_vars(CI='0')
    def test_include_check_partials_false_gl(self):
        dm.options['include_check_partials'] = False
        p = brachistochrone_min_time(transcription='gauss-lobatto', compressed=False,
                                     run_driver=False, force_alloc_complex=True)
        cpd = p.check_partials(out_stream=None, method='fd')
        self.assertSetEqual(set(cpd.keys()), {'traj0.phases.phase0.rhs_disc',
                                              'traj0.phases.phase0.rhs_col'})

    def test_include_check_partials_true_radau(self):
        dm.options['include_check_partials'] = True
        p = brachistochrone_min_time(transcription='radau-ps', compressed=False,
                                     run_driver=False, force_alloc_complex=True)
        cpd = p.check_partials(out_stream=None, method='fd')
        self.assertTrue(len(list(cpd.keys())) > 1)

    def test_include_check_partials_true_gl(self):
        dm.options['include_check_partials'] = True
        p = brachistochrone_min_time(transcription='gauss-lobatto', compressed=False,
                                     run_driver=False, force_alloc_complex=True)
        cpd = p.check_partials(out_stream=None, method='fd')
        self.assertTrue(len(list(cpd.keys())) > 1)



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal

import dymos as dm


class MainPhase(dm.Phase):

    def initialize(self):
        super(MainPhase, self).initialize()

    def setup(self):
        self.options['ode_class'] = TestODE
        super(MainPhase, self).setup()


class TestODE(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        self.add_subsystem('Sink',
                           om.ExecComp('sink = chord[0] + chord[1] + chord[2] + chord[3]',
                                       sink={'val': 0.0, 'units': None},
                                       chord={'val': np.zeros(4), 'units': 'm',
                                              'tags': ['dymos.static_target']}),
                           promotes_inputs=['chord'])

        self.add_subsystem('calc', om.ExecComp('Out = Thrust * 2',
                                               Out={'val': np.zeros(nn), 'units': 'N'},
                                               Thrust={'val': np.zeros(nn), 'units': 'N'}),
                           promotes_inputs=['Thrust'],
                           promotes_outputs=['Out'])


@use_tempdirs
class TestSimulateShapedParams(unittest.TestCase):

    def test_shaped_params(self):

        main_tx = dm.Radau(num_segments=1, order=3, compressed=False)

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        des_vars = p.model.add_subsystem('des_vars', om.IndepVarComp(), promotes_outputs=['*'])
        des_vars.add_output('chord', val=4 * np.ones(4), units='inch')

        hop0 = dm.Trajectory()
        p.model.add_subsystem('hop0', hop0)
        main_phase = hop0.add_phase(name='main_phase',
                                    phase=MainPhase(transcription=main_tx))

        main_phase.set_time_options(fix_initial=True, fix_duration=True, units='s')

        main_phase.add_parameter('chord', units='inch')
        p.model.connect('chord', 'hop0.main_phase.parameters:chord')

        main_phase.add_state('impulse', fix_initial=True, fix_final=False, units='N*s',
                             rate_source='Out',
                             solve_segments=False)

        main_phase.add_control('Thrust', units='N',
                               targets='Thrust',
                               lower=-3450, upper=-500,
                               order=5, opt=True, control_type='polynomial')

        main_phase.add_objective('impulse', loc='final', ref=-1)

        p.setup(mode='auto', check=['unconnected_inputs'], force_alloc_complex=True)

        p.set_val('hop0.main_phase.t_initial', 0.0)
        p.set_val('hop0.main_phase.t_duration', 10)
        p.set_val('hop0.main_phase.controls:Thrust', val=-3400, indices=om.slicer[:, 0])
        p.set_val('hop0.main_phase.states:impulse',  main_phase.interp('impulse', [0, 0]))

        p.run_driver()

        assert_near_equal(p.get_val('hop0.main_phase.timeseries.impulse')[-1, 0], -7836.66666, tolerance=1.0E-4)

        try:
            hop0.simulate()
        except Exception:
            self.fail('Simulate did not correctly complete.')

    def test_shaped_traj_params(self):

        main_tx = dm.Radau(num_segments=1, order=3, compressed=False)

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        hop0 = dm.Trajectory()
        p.model.add_subsystem('hop0', hop0)
        main_phase = hop0.add_phase(name='main_phase',
                                    phase=MainPhase(transcription=main_tx))

        main_phase.set_time_options(fix_initial=True, fix_duration=True, units='s')

        hop0.add_parameter('chord', units='inch', targets={'main_phase': ['chord']})

        main_phase.add_state('impulse', fix_initial=True, fix_final=False, units='N*s',
                             rate_source='Out',
                             solve_segments=False)

        main_phase.add_control('Thrust', units='N',
                               targets='Thrust',
                               lower=-3450, upper=-500,
                               order=5, opt=True, control_type='polynomial')

        main_phase.add_objective('impulse', loc='final', ref=-1)

        p.setup(mode='auto', check=['unconnected_inputs'], force_alloc_complex=True)

        p.set_val('hop0.main_phase.t_initial', 0.0)
        p.set_val('hop0.main_phase.t_duration', 10)
        p.set_val('hop0.main_phase.controls:Thrust', val=-3400, indices=om.slicer[:, 0])
        p.set_val('hop0.main_phase.states:impulse',  main_phase.interp('impulse', [0, 0]))

        p.run_driver()

        assert_near_equal(p.get_val('hop0.main_phase.timeseries.impulse')[-1, 0], -7836.66666, tolerance=1.0E-4)

        try:
            hop0.simulate()
        except Exception:
            self.fail('Simulate did not correctly complete.')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()
